#include <Alert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "TCPNetThread.h"
#include "NetException.h"
#include "Say.h"

//---------------------------------------------------------------------

TCPNetThread::TCPNetThread( char* name )
			  : NetThreadBase( name )
{
	read_thread = 0;
	m_socket = -1;
	
	Run();				// Start the BLooper's message thread
	
}


//---------------------------------------------------------------------

// BLooper's delete themselves, so cleanup here instead of the destructor
void TCPNetThread::Close() {

	Disconnect();
	KillReadThread();

	// Now tell the BLooper to kill itself
	PostMessage( new BMessage(B_QUIT_REQUESTED) );
}


//---------------------------------------------------------------------

bool TCPNetThread::MakeReadThread() {

	// Attempt to spawn the socket reader thread
	read_thread = spawn_thread( socket_read_entry, "read_thread", B_LOW_PRIORITY, this );
	
	// Make sure it got spawned all right
	if( read_thread <= 0 ) {
		HandleError( TCP_INTERNALERROR, "read_thread not spawned!" );
		return false;
	}
	
	return true;
}

//---------------------------------------------------------------------

void TCPNetThread::KillReadThread() {

	// kill the socket reader thread, if it's still alive
	if( read_thread )
		kill_thread( read_thread );
		
	read_thread = 0;
}

//---------------------------------------------------------------------

void TCPNetThread::WakeReadThread() {

	thread_info info;
	if( get_thread_info(read_thread, &info) == B_NO_ERROR ) {
		if( info.state == B_THREAD_SUSPENDED ) {
			resume_thread(read_thread);
		}
	}
}

//---------------------------------------------------------------------

// this has its own thread to avoid blocking everything while waiting for data.
// When data arrives, it notifies the BLooper and puts itself to sleep. It will
// be woken again (by other parts of the class) when we are waiting for more data.
int32 TCPNetThread::socket_read_entry( void *arg ) {

	TCPNetThread* this_class = (TCPNetThread*)arg;
	return this_class->do_socket_read();
}

//---------------------------------------------------------------------

int32 TCPNetThread::do_socket_read() {

	int retVal = WF_CONTINUE;
		
	// Loop until disconnect: block until a message arrives, notify the
	// net_thread that data is here, and wait for another message
	while( retVal == WF_GOTDATA || retVal == WF_CONTINUE ) {	

		retVal = socket_waitfor_data();

		// socket_waitfor_data() also detects when the connection has died
		if( retVal == WF_CONNDROPPED ) {
			PostMessage( BEAIM_NET_LOST_CONNECTION );			
			return 0;
		}

		// if the read thread was just interrupted, keep going
		if( retVal == WF_CONTINUE )
			continue;

		// if there is data waiting, tell the main thread and go to sleep
		if( retVal == WF_GOTDATA ) {
			PostMessage( BEAIM_NET_DATA_AVAILABLE );
			suspend_thread( read_thread );
		}
	}

	return 0;
}

//---------------------------------------------------------------------

// Returns true if data arrives, and false on error (connection dropped)
int TCPNetThread::socket_waitfor_data() {
	
	int retval, err;

	// receive 0 bytes of data (block until something's available)
	retval = ::recv( m_socket, NULL, 0, 0 );

	if( retval < 0 ) {
	
		err = errno;		
			
		// handle dead connection...
		if (err == EBADF || err == ENOTCONN || err == ECONNRESET) {
			return WF_CONNDROPPED;
		}
		
		// read() returned an error, but the connection is still good.
		// Probably the net thread wrote into this socket. There probably is a constant
		// reflecting this, but I couldn't find it so I will pretend that is what happened
		// whenever there is an error but the connection wasn't dropped. Fix later.
		else {
			//printf( "\n******* read thread wrote in, continuing...\n" );
			return WF_CONTINUE;
		}

	}
	
	return WF_GOTDATA;
}

//---------------------------------------------------------------------

bool TCPNetThread::MakeConnectThread() {

	printf( "gonna spawn connect thread\n" );

	// Attempt to spawn the socket reader thread
	connect_thread = spawn_thread( connect_entry, "connect_thread", B_LOW_PRIORITY, this );
	
	// Make sure it got spawned all right
	if( connect_thread <= 0 ) {
		printf( "Error - connect_thread not spawned!\n" );
		return false;
	}
	
	// Now wake it up... it has a job to do right away
	resume_thread( connect_thread );
	
	return true;
}

//---------------------------------------------------------------------

void TCPNetThread::KillConnectThread() {

	// kill the connect thread, if it's still alive
	if( connect_thread )
		kill_thread( connect_thread );
		
	connect_thread = 0;
}


//---------------------------------------------------------------------

int32 TCPNetThread::connect_entry( void *arg ) {

	printf( "connect entry\n" );
	fflush(stdout);

	TCPNetThread* this_class = (TCPNetThread*)arg;
	return this_class->do_connect();
}

//---------------------------------------------------------------------

int32 TCPNetThread::do_connect() {

	// try to connect
	try {

		printf( "connect to: %s (%ld)\n", conHost, conPort );
		socket_connect( conHost, conPort );
	}
	catch (NetException &ex)
	{
		HandleError( TCP_COULDNOTCONNECT );
		return 0;
	}
	
	printf( "connected!!!!!!!!!!!!!!!!!!!\n" );
	
	// all connected... send a message that says so.
	PostMessage( new BMessage(BEAIM_NET_CONNECTED) );

	// NULL out this thread, and return
	connect_thread = 0;
	return 1;
}

//---------------------------------------------------------------------

// This is, after all, a BLooper... gotta have a MessageReceived!
void TCPNetThread::MessageReceived( BMessage *msg )
{
	switch( msg->what ) {

		case BEAIM_NET_CONNECTED:
			printf( "this sucker is connected!\n" );
			nowConnected();
			break;
		
		case BEAIM_NET_DATA_AVAILABLE:
			DataWaiting();
			WakeReadThread();
			break;
			
		case BEAIM_NET_LOST_CONNECTION:
			Disconnected();
			break;
	};
}

//---------------------------------------------------------------------

// default connect function - can be overridden
bool TCPNetThread::Connect( char* host, int32 port, bool fs ) {

	printf( "connector.\n" );

	// copy in the connection data
	strcpy( conHost, host );
	conPort = port;
	fullServiceConnection = fs;

	// Set up the connection
	printf( "connect called\n" );
	MakeConnectThread();
	
	return true;		
}

//---------------------------------------------------------------------

void TCPNetThread::nowConnected() {

	// call the connect handler
	if( fullServiceConnection ) {
		MakeReadThread();
		WakeReadThread();
		Connected();
	}
}

//---------------------------------------------------------------------

// default disconnect function - can be overridden
void TCPNetThread::Disconnect() {
	
	status_t exitcode;
	
	// Kill the connect thread (in case we never got connected)
	KillConnectThread();
	
	// Now close the socket, which will trip the read thread, and wait for
	//  the read thread to close before returning
	socket_close();
	wait_for_thread( read_thread, &exitcode );
	
	// Make the socket invalid, just in case
	m_socket = -1;
}

//---------------------------------------------------------------------

void TCPNetThread::socket_reset() {

	char errbuf[256];
	socket_close();
	m_socket = ::socket( AF_INET, SOCK_STREAM, 0 );
	if(m_socket < 0)
	{
		m_state = s_inval;
		sprintf(errbuf, "::socket() failed: %s", strerror(errno));
		NET_THROW_ARG("socket_reset()", errbuf, this);
	}
	m_state |= ::s_open;
}

//---------------------------------------------------------------------

void TCPNetThread::socket_close() {

	if(m_state & ::s_open)
	{
		closesocket(m_socket);
		m_socket = -1;
	}
	m_state = s_inval;
}

//---------------------------------------------------------------------

void TCPNetThread::socket_connect( char* hostname, int port ) {

	struct sockaddr_in sa;
	struct hostent *he;
	in_addr ia;	
	char errbuf[256];
	int s = sizeof(struct sockaddr_in);
	
	if((m_state & ::s_open) != true)
		socket_reset();
	
	ia.s_addr = INADDR_ANY;
	if( hostname != 0 )
	{
		ia.s_addr = inet_addr(hostname);
		if(ia.s_addr == INADDR_ANY || ia.s_addr == (unsigned long) -1)
		{
			he = ::gethostbyname(hostname);
			if(he != 0)
			{
				ia.s_addr = *(int *)he->h_addr_list[0];
			} else {
				char errbuf[256];
				sprintf(errbuf, "::gethostbyname failed for host name %s: %s",
					hostname, strerror(h_errno));
				NET_THROW("NLAddress::set", errbuf);
			}
		}
	}
	memset(&sa, 0, sizeof(sa));
	sa.sin_family = AF_INET;
	sa.sin_addr = ia;
	sa.sin_port = htons(port);	

	if(::connect(m_socket, (struct sockaddr *) &sa, s) < 0)
	{
		socket_close();
		sprintf(errbuf, "::connect() failed: %s", strerror(errno));
		NET_THROW_ARG("NLEndpoint::connect()", errbuf, this);	
	}
	
	if(::getpeername(m_socket, (struct sockaddr *) &sa, &s) < 0)
	{
		socket_close();
		sprintf(errbuf, "::getpeername() failed: %s", strerror(errno));
		NET_THROW_ARG("NLEndpoint::connect()", errbuf, this);	
	}
	
	m_state |= ::s_connected;
}

//---------------------------------------------------------------------

ssize_t TCPNetThread::socket_send( const void *buf, size_t size ) {

	ssize_t rc = 0;

	//printf( "@@@@@@@@@@@@@@socket sending...\n" );
	
	if( m_state & ::s_connected )
		rc = ::send( m_socket, (const char *) buf, size, 0 );

	if( rc < 0 )
	{
		//if( m_throwsend == true )
		{
			char errbuf[256];
			sprintf(errbuf, "::send() failed: %s", strerror(errno));
			NET_THROW_ARG("NLEndpoint::send()", errbuf, this);	
		}
	}
	
	//printf( "@@@@@@@@@@@@@@socket sent!!\n" );
		
	return rc;
}

//---------------------------------------------------------------------

ssize_t TCPNetThread::socket_recv( void *buf, size_t size ) {

	long rc;
	
	//if(m_timeout >= 0 && isDataPending(m_timeout) == false)
	//	return 0;

	//printf( "@@@@@@@@@@@@@@socket recv'ing...\n" );
		
	rc = ::recv( m_socket, (char *) buf, size, 0 );
	if(rc < 0)
	{
		//if(m_throwrecv == true)
		{
			char errbuf[256];
			sprintf(errbuf, "::recv() failed: %s", strerror(errno));
			NET_THROW_ARG("NLEndpoint::recv()", errbuf, this);	
		}
	}
	
	//printf( "@@@@@@@@@@@@@@socket recv'd!!\n" );
	
	return rc;
}

//---------------------------------------------------------------------

/* Login: STANDARD BMESSAGE DATA:
	userid - string - the userid, or screen name, or whatever.
	password - string - the password.   */
void TCPNetThread::Login( BMessage* ) {}


/* Logout: STANDARD BMESSAGE DATA:
	empty.   */
void TCPNetThread::Logout( BMessage* ) {}

// Virtual handler functions
void TCPNetThread::DataWaiting() {}
void TCPNetThread::Connected() {}
void TCPNetThread::Disconnected() {}
void TCPNetThread::HandleError( int error, const char* message ) {}
